import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.LinkedHashMap; 
import controlP5.*; 
import controlP5.Controller; 
import java.util.Map; 
import java.util.Map.Entry; 
import java.lang.reflect.Field; 
import java.util.Arrays; 
import net.java.games.input.*; 
import org.gamecontrolplus.*; 
import java.util.Map.Entry; 
import java.util.Vector; 
import java.lang.reflect.Field; 
import java.lang.reflect.Modifier; 
import java.util.Arrays; 
import java.util.List; 
import java.util.Set; 
import java.time.OffsetDateTime; 
import java.util.Collections; 
import java.util.Comparator; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class jumplab extends PApplet {

/*
 * JumpLab: a testbed for studying jump motion and scrolling algorithms of videogames
 *
 * Web: https://fukuchi.org/works/jumplab/
 * Repository: https://github.com/fukuchi/jumplab/
 *
 * Copyright (C) 2020, 2021 Kentaro Fukuchi <kentaro@fukuchi.org> and other contributors
 *
 * This program is free software: you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program.
 * If not, see https://www.gnu.org/licenses/.
 */

Settings gSettings;
Jumper gMasao;
Level gLevel;
Camera gCamera;
Console gConsole;
Joystick gJoystick;
StyleManager gStyles;

boolean gPause;
boolean gStepForward;

static final String gVersionString = "1.3.0";

static final int gameScreen_w = 800;
static final int gameScreen_h = 600;
static final int console_w = 300;

static final String userSettingsFilename = "user_settings.json";
static final String defaultSettingsFilename = "default_settings.json";
static final String joystickConfigFilename = "joystick_config.json";

public void settings() {
  println("JumpLab version " + gVersionString);
  println("Initializing...");
  // If you have any performance problem, try P2D render as follows.
  size(gameScreen_w + console_w, gameScreen_h);
  //size(gameScreen_w + console_w, gameScreen_h, P2D);
  noSmooth();
}

public void setup() {
  gStyles = new StyleManager();
  if (!gStyles.loadDefaultSettings(defaultSettingsFilename)) {
    System.err.println("Failed to load the default style settings. The installed package seems to be broken. Check the files under the installed directory.");
    exit();
  }
  gStyles.loadUserSettings(userSettingsFilename);

  gSettings = new Settings();
  gJoystick = new Joystick(this, joystickConfigFilename);
  gLevel = new Level("level1.csv", "tiles.png", "bg.png");
  gMasao = new Jumper(gSettings, gLevel);
  gCamera = new Camera(gMasao, gLevel, gSettings, gameScreen_w, gameScreen_h);
  gCamera.reset(gMasao.center_x(), gMasao.center_y());
  gConsole = new Console(this, gameScreen_w, 0, console_w, gameScreen_h, gStyles, gSettings);
  frameRate(60);

  background(128);
  println("Initialization completed.");
}

public void draw() {
  if (!gPause || gStepForward) {
    gMasao.update();
    gCamera.update();
    gConsole.statusUpdate(gMasao, gLevel, gCamera);
    gStepForward = false;
  } else {
    gJoystick.updateDuringPause();
  }
  gCamera.draw();
  gConsole.drawStatus(gMasao);
}

public void keyPressed() {
  boolean eventReceived;

  eventReceived = gMasao.keyPressed();
  if (!eventReceived) gConsole.keyPressed();
}

public void keyReleased() {
  boolean eventReceived;

  eventReceived = gMasao.keyReleased();
  if (!eventReceived) gConsole.keyReleased();
}


enum ButtonFunction {
  NONE("None"), JUMP("Jump"), DASH("Dash"), TOGGLE_TRAIL("Toggle show Trail"), TOGGLE_CAMERA("Toggle show Camera Marker"), 
    NEXT_STYLE("Next style"), PREV_STYLE("Previous style"), PAUSE("Pause"), STEP_FORWARD("Step forward");

  private final String label;

  private ButtonFunction(final String label) {
    this.label = label;
  }

  public String getLabel() {
    return label;
  }

  private static final Map<String, Object> valuesMap;
  private static final ButtonFunction[] reverseLookupTable;
  private static final boolean[] enabledWhilePausing;
  private static final int size = ButtonFunction.values().length;
  static {
    valuesMap = new LinkedHashMap<String, Object>();
    for (ButtonFunction func : ButtonFunction.values()) {
      valuesMap.put(func.getLabel(), func);
    }
    reverseLookupTable = ButtonFunction.values();
    enabledWhilePausing = new boolean[size];
    enabledWhilePausing[PAUSE.ordinal()] = true;
    enabledWhilePausing[STEP_FORWARD.ordinal()] = true;
  }

  public static Map<String, Object> getValuesMap() {
    return valuesMap;
  }

  public static ButtonFunction getEnumByOrdinal(int o) {
    return reverseLookupTable[o];
  }

  public boolean isEnabledWhilePausing() {
    return enabledWhilePausing[ordinal()];
  }

  public static boolean isEnabledWhilePausing(int o) {
    return enabledWhilePausing[o];
  }
}
class Camera {
  Jumper jumper;
  Level level;
  Settings settings;
  PImage levelImg;
  PImage titleImg;
  JumperTrail[] trail;
  static final int trailLen = 64;
  int trailHead = 0;
  boolean showTitle = true;
  int titleTimer;

  float x, y;
  float px, py; // previous position
  float focus_x;
  float targetFocus_x;
  int window_w, window_h;
  int window_hw, window_hh;

  PFont onScreenFont;

  class JumperTrail {
    float x, y;
    PImage image;
  }

  Camera(Jumper jumper, Level level, Settings settings, int w, int h) {
    this.jumper = jumper;
    this.level = level;
    this.settings = settings;
    window_w = w;
    window_h = h;
    window_hw = w / 2;
    window_hh = h / 2;

    levelImg = level.createLevelImage();
    titleImg = loadImage("title.png");
    titleTimer = -1;

    trail = new JumperTrail[trailLen];
    for (int i=0; i<trailLen; i++) {
      trail[i] = new JumperTrail();
    }

    onScreenFont = createFont("Lucida Sans", 16);
  }

  public void reset(float x, float y) {
    this.x = constrain(x, window_hw, level.w - window_hw);
    this.y = constrain(y, window_hh, level.h - window_hh);
    px = this.x;
    py = this.y;
    focus_x = 0;
  }

  public void update() {
    float tx = x; // target position
    float ty = y;

    float dx = jumper.center_x() - x;
    float dy = jumper.center_y() - y;
    if (settings.forwardFocus || settings.projectedFocus) {
      targetFocus_x = 0;
      if (settings.forwardFocus) {
        targetFocus_x += jumper.lastDir * settings.focusDistance;
      }
      if (settings.projectedFocus) {
        targetFocus_x += settings.focusDistance * jumper.vx / settings.maxVx;
      }
      if (jumper.lastDir > 0) {
        if (focus_x < targetFocus_x) {
          float cameraEdge_x = settings.cameraWindow_w / 2 - dx;
          if (focus_x < cameraEdge_x) {
            focus_x = cameraEdge_x;
          }
          focus_x += settings.focusingSpeed;
          if (focus_x > targetFocus_x) {
            focus_x = targetFocus_x;
          }
        } else if (focus_x > targetFocus_x) {
          focus_x -= 1;
          if (focus_x < targetFocus_x) {
            focus_x = targetFocus_x;
          }
        }
      } else if (jumper.lastDir < 0) {
        if (focus_x > targetFocus_x) {
          float cameraEdge_x = -settings.cameraWindow_w / 2 - dx;
          if (focus_x > cameraEdge_x) {
            focus_x = cameraEdge_x;
          }
          focus_x -= settings.focusingSpeed;
          if (focus_x < targetFocus_x) {
            focus_x = targetFocus_x;
          }
        } else if (focus_x < targetFocus_x) {
          focus_x += 1;
          if (focus_x > targetFocus_x) {
            focus_x = targetFocus_x;
          }
        }
      }
      dx += focus_x;
    }
    if (dx < -settings.cameraWindow_w / 2) {
      tx += dx + settings.cameraWindow_w / 2;
    } else if (dx > settings.cameraWindow_w / 2) {
      tx += dx - settings.cameraWindow_w / 2;
    }
    if (dy < -settings.cameraWindow_h / 2) {
      ty += dy + settings.cameraWindow_h / 2;
    } else if (dy > settings.cameraWindow_h / 2) {
      ty += dy - settings.cameraWindow_h / 2;
    }

    if (!settings.cameraEasing_x) {
      px = x;
      x = tx;
    } else {
      float vx = x - px;
      px = x;
      x += vx * (0.5f - settings.cameraEasingNormal_x) + (tx - x) * settings.cameraEasingNormal_x;
    }
    x = constrain(x, window_hw, level.w - window_hw);

    if (!settings.platformSnapping || !jumper.jumping) {
      if (!settings.cameraEasing_y) {
        py = y;
        y = ty;
      } else {
        float vy = y - py;
        py = y;
        if (!jumper.onObstacle) {
          y += vy * (0.5f - settings.cameraEasingNormal_y) + (ty - y) * settings.cameraEasingNormal_y;
        } else {
          y += vy * (0.5f - settings.cameraEasingGrounding_y) + (ty - y) * settings.cameraEasingGrounding_y;
        }
      }
      y = constrain(y, window_hh, level.h - window_hh);
    }
  }

  public void draw() {
    int bgx, bgy;
    int cx = (int)x - window_hw;
    int cy = (int)y - window_hh;
    float rx = (float)cx / (level.w - window_w);
    float ry = (float)cy / (level.h - window_h);
    bgx = (int)((level.bgImg.width  - window_w) * (0.5f + settings.bgScrollRatio * (rx - 0.5f)));
    bgy = (int)((level.bgImg.height - window_h) * (0.7f + settings.bgScrollRatio * (ry - 0.7f)));

    copy(level.bgImg, bgx, bgy, window_w, window_h, 0, 0, window_w, window_h);
    copy(levelImg, cx, cy, window_w, window_h, 0, 0, window_w, window_h);

    if (settings.showTrail) drawTrail(cx, cy);
    PImage img = jumper.draw(cx, cy);
    if (!gPause || gStepForward) {
      trail[trailHead].x = jumper.x;
      trail[trailHead].y = jumper.y;
      trail[trailHead].image = img;
      trailHead++;
      if (trailHead >= trailLen) trailHead = 0;
    }

    if (settings.showCameraMarker) drawCameraMarker(cx, cy);
    if (showTitle) drawTitle();
    if (settings.showInputStatus) drawInputStatus();
    if (gPause) drawPauseBar();

    noStroke();
    fill(128);
    rect(gConsole.x, gConsole.y, gConsole.w, gConsole.h);
    stroke(255);
    line(gConsole.x, gConsole.y + 166, gConsole.x + gConsole.w, gConsole.y + 166);
  }

  public void drawTrail(int cx, int cy) {
    pushStyle();
    if (!settings.showAfterimage) {
      noStroke();
      fill(255, 0, 0);
      for (int i=0; i<trailLen; i++) {
        ellipse(trail[i].x + Jumper.w / 2 - cx, trail[i].y + Jumper.h / 2 - cy, 5, 5);
      }
    } else {
      tint(255, 128);
      for (int i=0; i<trailLen; i++) {
        int idx = (trailHead + 1 + i) % trailLen;
        if (idx % 4 == 0) {
          jumper.draw(trail[idx].image, trail[idx].x - cx, trail[idx].y - cy);
        }
      }
    }
    popStyle();
  }

  public void drawCameraMarker(int cx, int cy) {
    pushStyle();
    noStroke();
    fill(255, 0, 0);
    rect(window_hw -  2, window_hh - 12, 5, 25);
    rect(window_hw - 12, window_hh -  2, 25, 5);
    stroke(255);
    noFill();
    rect(window_hw - settings.cameraWindow_w / 2, window_hh - settings.cameraWindow_h / 2, settings.cameraWindow_w, settings.cameraWindow_h);
    if (settings.forwardFocus || settings.projectedFocus) {
      int fx = (int)(jumper.center_x() + targetFocus_x) - cx;
      int fy = (int)jumper.center_y() - cy;
      line(fx - 9, fy - 9, fx + 9, fy + 9);
      line(fx - 9, fy + 9, fx + 9, fy - 9);
      stroke(255, 255, 0);
      fx = (int)(jumper.center_x() + focus_x) - cx;
      fy = (int)jumper.center_y() - cy;
      line(fx - 12, fy, fx + 12, fy     );
      line(fx, fy - 12, fx, fy + 12);
    }
    popStyle();
  }

  public void drawInputStatus() {
    pushMatrix();
    pushStyle();
    translate(window_hw - 170, 40);
    stroke(0);
    strokeWeight(3);
    if (jumper.inputStatus[0]) {
      fill(64, 255, 255);
    } else {
      fill(0, 64, 64);
    }
    triangle(0, 40, 60, 0, 60, 80);
    translate(80, 0);
    if (jumper.inputStatus[1]) {
      fill(64, 255, 255);
    } else {
      fill(0, 64, 64);
    }
    triangle(20, 0, 80, 40, 20, 80);
    translate(100, 0);
    if (jumper.inputStatus[2]) {
      fill(255, 64, 64);
    } else {
      fill(64, 0, 0);
    }
    ellipse(40, 40, 70, 70);
    if (jumper.inputStatus[3]) {
      fill(255, 64, 64);
    } else {
      fill(64, 0, 0);
    }
    ellipse(120, 40, 70, 70);
    popMatrix();
    popStyle();
  }

  public void drawTitle() {
    pushStyle();
    int now = millis();
    if (titleTimer < 0) {
      titleTimer = now + 3500;
    } else if (now < titleTimer) {
      int remaining = titleTimer - now;
      if (remaining >= 256 && (jumper.vx != 0 || !jumper.onObstacle)) {
        titleTimer -= 100;
      }
      if (remaining < 256) {
        tint(255, remaining);
      }
      image(titleImg, (window_w - titleImg.width) / 2, 50);
      textAlign(RIGHT, CENTER);
      textFont(onScreenFont);
      fill(255, remaining);
      text("Version " + gVersionString, (window_w - titleImg.width) / 2 + titleImg.width, 50 + titleImg.height);
    } else {
      showTitle = false;
    }
    popStyle();
  }

  public void drawPauseBar() {
    pushMatrix();
    pushStyle();
    translate(window_hw, window_hh - 50);
    textFont(onScreenFont);
    textSize(16);
    float w = textWidth("PAUSE");
    noStroke();
    fill(0);
    rect(-w / 2 - 50, -10, w + 100, 20);
    fill(255);
    text("PAUSE", -w / 2, 7);
    popMatrix();
    popStyle();
  }
}
class ChartCanvas extends Canvas {
  int x, y;
  PGraphics pg;
  ArrayList<Series> serieses;
  int seriesesNum;
  Settings settings;

  class Series {
    String label;
    float[] data;

    Series(String label) {
      this.label = label;
      data = new float[3];
    }
  }

  ChartCanvas(Settings settings, int x, int y, int w, int h) {
    this.x = x;
    this.y = y;
    this.settings = settings;
    pg = createGraphics(w, h);
    pg.beginDraw();
    pg.background(255);
    pg.endDraw();
    serieses = new ArrayList<Series>();
    seriesesNum = 0;
  }

  public void draw(PGraphics destPg) {
    destPg.pushMatrix();
    destPg.pushStyle();
    destPg.translate(x, y);
    destPg.image(pg, 0, 0);
    destPg.translate(0, pg.height + 10);
    destPg.colorMode(HSB, 360, 100, 100);
    destPg.textSize(10);
    destPg.textAlign(LEFT, CENTER);
    for (int i=0; i<seriesesNum; i++) {
      destPg.fill(0, 0, 100);
      destPg.text(serieses.get(i).label, 200, i * 20);
      destPg.fill(360 * i / seriesesNum, 100, 100);
      destPg.noStroke();
      destPg.rect(150, i * 20 - 1, 40, 3);
    }
    destPg.popStyle();
    destPg.popMatrix();
  }

  public int addSeries(String label) {
    Series s = new Series(label);
    serieses.add(seriesesNum, s);
    seriesesNum++;
    return seriesesNum - 1;
  }

  public void updateSeries(int idx, float value) {
    float[] series = serieses.get(idx).data;
    series[2] = series[1];
    series[1] = series[0];
    series[0] = value;
  }

  public void showVelocityChartChanged() {
    pg.beginDraw();
    pg.background(255);
    pg.endDraw();
  }

  public void updateChart() {
    pg.beginDraw();
    pg.pushStyle();
    pg.copy(1, 0, pg.width - 1, pg.height, 0, 0, pg.width - 1, pg.height);
    pg.colorMode(RGB, 255);
    pg.fill(255);
    pg.noStroke();
    pg.rect(pg.width - 1, 0, 1, pg.height);
    pg.colorMode(HSB, 360, 100, 100);
    if (!settings.showVelocityChart) {
      for (int i=0; i<seriesesNum; i++) {
        pg.stroke(360 * i / seriesesNum, 100, 100);
        float[] series = serieses.get(i).data;
        pg.line(pg.width - 2, (pg.height - 1) * (1 - series[1]), pg.width - 1, (pg.height - 1) * (1 - series[0]));
      }
    } else {
      float pg_hh = pg.height / 2;
      pg.stroke(0, 0, 75);
      pg.line(pg.width - 2, pg_hh, pg.width - 1, pg_hh);
      for (int i=0; i<seriesesNum; i++) {
        pg.stroke(360 * i / seriesesNum, 100, 100);
        float[] series = serieses.get(i).data;
        float d1 = pg_hh - (series[1] - series[2]) * 10 * (pg.height - 1);
        float d2 = pg_hh - (series[0] - series[1]) * 10 * (pg.height - 1);
        pg.line(pg.width - 2, d1, pg.width - 1, d2);
      }
    }
    pg.popStyle();
    pg.endDraw();
  }
}
class ComparableVersion implements Comparable<ComparableVersion> {
  private int[] numbers;
  String versionString;

  ComparableVersion(String versionString) {
    this.versionString = versionString;
    String[] strs = versionString.split("\\.", 3);
    numbers = new int[strs.length];

    for (int i=0; i<strs.length; i++) {
      numbers[i] = Integer.parseInt(strs[i]);
    }
  }

  public int number(int level) {
    if (level >= numbers.length) {
      return -1;
    }
    return numbers[level];
  }

  public int compareTo(ComparableVersion other, int level) {
    int thisNum = this.number(level);
    int otherNum = other.number(level);
    int res = Integer.compare(thisNum, otherNum);
    if (res == 0) {
      if (thisNum >= 0 && otherNum >= 0) {
        res = compareTo(other, level + 1);
      }
    }
    return res;
  }

  public int compareTo(ComparableVersion other) {
    return compareTo(other, 0);
  }

  public int compareTo(String other) {
    return compareTo(new ComparableVersion(other));
  }

  public String toString() {
    return versionString;
  }
}







class Console {
  int x, y, w, h;
  ControlP5 ctlr;
  HashMap<String, Controller> widgets;
  HashMap<Integer, String> id2parameter;
  HashMap<String, int[]> indicators;
  HashMap<String, Integer> buttonFunctionSelectorsMap;
  StyleManager styles;
  Settings settings;
  boolean halfFilled = false;
  Controller lastWidget;
  int nextWidgetPosition_y;
  static final int widgetMargin_y = 6;
  NumIndicator numIndicator;
  ChartCanvas chart;
  int chartJumperXIdx;
  int chartJumperYIdx;
  int chartCameraXIdx;
  int chartCameraYIdx;
  String currentTab = "global";
  CColor scrollableListItemColor = new CColor().setBackground(color(32, 64, 192));

  Console(PApplet parent, int x, int y, int w, int h, StyleManager styles, Settings settings) {
    this.x = x;
    this.y = y;
    this.w = w;
    this.h = h;
    this.styles = styles;
    this.settings = settings;

    ctlr = new ControlP5(parent);
    widgets = new HashMap<String, Controller>();
    id2parameter = new HashMap<Integer, String>();
    indicators = new HashMap<String, int[]>();
    buttonFunctionSelectorsMap = new HashMap<String, Integer>();

    ctlr.getTab("default").setLabel("Jump1");
    ctlr.addTab("Jump2");
    ctlr.addTab("Camera");
    ctlr.addTab("Character");
    ctlr.addTab("Joystick");
    ctlr.addTab("Chart");
    ctlr.addTab("Misc");
    ctlr.getWindow().setPositionOfTabs(x + 10, y + 150);

    stroke(255);
    ctlr.addTextlabel("ProgramTitle")
      .setPosition(x + 10, y + 5)
      .setText("JumpLab ver. " + gVersionString)
      .moveTo("global");
    ctlr.addTextlabel("JumperXY")
      .setPosition(x + 10, y + 25)
      .setText("Jumper X,Y")
      .moveTo("global");
    indicators.put("JumperXYvalue", new int[] {x + 80, y + 25});
    ctlr.addTextlabel("JumperVXY")
      .setPosition(x + 10, y + 40)
      .setText("Jumper VX,VY")
      .moveTo("global");
    indicators.put("JumperVXYvalue", new int[] {x + 80, y + 40});
    ctlr.addTextlabel("JumperAXY")
      .setPosition(x + 10, y + 55)
      .setText("Jumper AX,AY")
      .moveTo("global");
    indicators.put("JumperAXYValue", new int[] {x + 80, y + 55});
    ctlr.addTextlabel("PropellingRemainingFrames")
      .setPosition(x + 10, y + 70)
      .setText("Propelling Remaining")
      .moveTo("global");
    indicators.put("PropellingRemainingFramesValue", new int[] {x + 98, y + 70});
    ctlr.addTextlabel("FPS")
      .setPosition(x + 200, y + 70)
      .setText("FPS")
      .moveTo("global");
    ctlr.addFrameRate().setInterval(10).setPosition(x + 260, y + 70).moveTo("global");
    ctlr.addTextlabel("Jumping")
      .setPosition(x + 200, y + 25)
      .setText("Jumping")
      .moveTo("global");
    widgets.put("JumpingValue", ctlr.addTextlabel("JumpingValue")
      .setPosition(x + 260, y + 25)
      .moveTo("global"));
    ctlr.addTextlabel("Propelling")
      .setPosition(x + 200, y + 40)
      .setText("Propelling")
      .moveTo("global");
    widgets.put("PropellingValue", ctlr.addTextlabel("PropellingValue")
      .setPosition(x + 260, y + 40)
      .moveTo("global"));
    ctlr.addTextlabel("OnObstacle")
      .setPosition(x + 200, y + 55)
      .setText("On obstacle")
      .moveTo("global");
    widgets.put("OnObstacleValue", ctlr.addTextlabel("OnObstracleValue")
      .setPosition(x + 260, y + 55)
      .moveTo("global"));

    ctlr.addTextlabel("Preset Label")
      .setPosition(x + 10, y + 95)
      .setText("STYLES")
      .moveTo("global");
    ScrollableList slist = ctlr.addScrollableList("Styles List")
      .setBackgroundColor(color(192))
      .setPosition(x + 60, y + 90)
      .setSize(150, 100)
      .setBarHeight(20)
      .setItemHeight(15)
      .setItems(styles.keyList())
      .moveTo("global")
      .plugTo(this, "styleSelected");
    slist.getValueLabel().toUpperCase(false);
    slist.getCaptionLabel().toUpperCase(false);
    setItemsColor(slist, scrollableListItemColor);

    Textfield textfield = ctlr.addTextfield("style name")
      .setPosition(x + 60, y + 120)
      .setSize(150, 20)
      .setAutoClear(false)
      .moveTo("global")
      .plugTo(this, "styleNameChanged");
    textfield.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, ControlP5.CENTER);
    textfield.getCaptionLabel().setPadding(5, 0);
    slist.getValueLabel().setFont(textfield.getValueLabel().getFont()); // workaround to get defaultFontForText
    slist.getCaptionLabel().setFont(textfield.getValueLabel().getFont());
    ctlr.addButton("Save")
      .setPosition(x + 220, y + 120)
      .setSize(60, 20)
      .moveTo("global");

    // Settings of jumping
    setTab("default");
    nextWidgetPosition_y = y + 175;
    appendHalfwidthWidget("showTrail", ctlr.addToggle("Show trail"));
    appendHalfwidthWidget("stopAndFall", ctlr.addToggle("Stop and fall"));
    appendHalfwidthWidget("allowAerialJump", ctlr.addToggle("Allow aerial jump"));
    appendHalfwidthWidget("allowAerialWalk", ctlr.addToggle("Allow aerial walk"));
    appendHalfwidthWidget("allowAerialTurn", ctlr.addToggle("Allow aerial turn"));
    appendHalfwidthWidget();
    appendHalfwidthWidget("maxVx", ctlr.addSlider("Max Vx")
      .setRange(1, 20));
    appendHalfwidthWidget("maxVy", ctlr.addSlider("Max Vy")
      .setRange(1, 80));
    appendFullwidthWidget("jumpVelocity", ctlr.addSlider("Jump Velocity")
      .setRange(1, 30));
    appendFullwidthWidget("jumpVelocityBonus", ctlr.addSlider("Jump Velocity Bonus")
      .setRange(0, 0.5f));
    appendFullwidthWidget("jumpAnticipationFrames", ctlr.addSlider("Jump Anticipation Frames")
      .setRange(0, 14)
      .setNumberOfTickMarks(15)
      .showTickMarks(false));
    appendFullwidthWidget("vxAdjustmentAtTakeoff", ctlr.addSlider("VX Adjustment at Takeoff")
      .setRange(-1.0f, 1.0f)
      .setSliderMode(Slider.FLEXIBLE));
    appendFullwidthWidget("maxPropellingFrames", ctlr.addSlider("Maximum Propelling Duration")
      .setRange(0, 100)
      .setNumberOfTickMarks(101)
      .showTickMarks(false));
    appendFullwidthWidget("gravity", ctlr.addSlider("Gravity (rising)")
      .setRange(0, 2));
    appendFullwidthWidget("gravityFalling", ctlr.addSlider("Gravity (falling)")
      .setRange(0.01f, 2));
    appendFullwidthWidget("verticalSpeedSustainLevel", ctlr.addSlider("Jump Speed Sustain Level")
      .setRange(0, 1));
    appendFullwidthWidget("axNormal", ctlr.addSlider("X Accel (normal)")
      .setRange(0, 3));
    appendFullwidthWidget("axBrake", ctlr.addSlider("X Accel (braking)")
      .setRange(0, 3));
    appendFullwidthWidget("axJumping", ctlr.addSlider("X Accel (jumping)")
      .setRange(0, 3));
    appendFullwidthWidget("collisionTolerance", ctlr.addSlider("Collision Tolerance")
      .setRange(0, 24)
      .setNumberOfTickMarks(25)
      .showTickMarks(false));

    setTab("Jump2");
    nextWidgetPosition_y = y + 175;
    appendHalfwidthWidget("allowWallJump", ctlr.addToggle("Allow wall jump"));
    appendHalfwidthWidget("allowWallSlide", ctlr.addToggle("Allow wall slide"));
    appendFullwidthWidget("wallJumpSpeedRatio", ctlr.addSlider("Wall jump speed ratio")
      .setRange(0, 1));
    appendHalfwidthWidget("maxVxDashing", ctlr.addSlider("Max Vx Dashing")
      .setRange(1, 20));
    appendFullwidthWidget("axNormalDashing", ctlr.addSlider("X Accel (normal, dashing)")
      .setRange(0, 3));
    appendFullwidthWidget("vCollisionTolerance", ctlr.addSlider("Vertical Collision Tolerance")
      .setRange(0, 24)
      .setNumberOfTickMarks(25)
      .showTickMarks(false));

    // Settings of Camera motion
    setTab("Camera");
    nextWidgetPosition_y = y + 175;
    appendHalfwidthWidget("showCameraMarker", ctlr.addToggle("Camera marker"));
    appendHalfwidthWidget("cameraEasing_x", ctlr.addToggle("Camera easing X"));
    appendHalfwidthWidget("forwardFocus", ctlr.addToggle("Forward focus"));
    appendHalfwidthWidget("cameraEasing_y", ctlr.addToggle("Camera easing Y"));
    appendHalfwidthWidget("projectedFocus", ctlr.addToggle("Projected focus"));
    appendHalfwidthWidget("platformSnapping", ctlr.addToggle("Platform snapping"));
    appendFullwidthWidget("cameraEasingNormal_x", ctlr.addSlider("Camera X Easing Coef (normal)")
      .setRange(0.01f, 0.5f));
    appendFullwidthWidget("cameraEasingNormal_y", ctlr.addSlider("Camera Y Easing Coef (normal)")
      .setRange(0.01f, 0.5f));
    appendFullwidthWidget("cameraEasingGrounding_y", ctlr.addSlider("Camera Y Easing Coef (grounding)")
      .setRange(0.01f, 0.5f));
    appendFullwidthWidget("cameraWindow_w", ctlr.addSlider("Camera Window Width")
      .setRange(0, 400));
    appendFullwidthWidget("cameraWindow_h", ctlr.addSlider("Camera Window Height")
      .setRange(0, 400));
    appendFullwidthWidget("focusDistance", ctlr.addSlider("Focus Distance")
      .setRange(0, 400));
    appendFullwidthWidget("focusingSpeed", ctlr.addSlider("Focusing Speed")
      .setRange(0, 30));
    appendFullwidthWidget("bgScrollRatio", ctlr.addSlider("BG Scroll Speed")
      .setRange(0, 1));

    // Settings of character
    setTab("Character");
    nextWidgetPosition_y = y + 175;
    appendFullwidthWidget(null, ctlr.addTextlabel("Select Character")
      .setText("SELECT CHARACTER:"));
    ScrollableList characterList = ctlr.addScrollableList("Character List")
      .setBackgroundColor(color(192))
      .setSize(280, 100)
      .setBarHeight(20)
      .setItemHeight(15)
      .setItems(gMasao.proportionLabels)
      .setValue(1)
      .plugTo(this, "characterSelected");
    appendFullwidthWidget("selectedCharacter", characterList);

    // Settings of joystick
    setTab("Joystick");
    nextWidgetPosition_y = y + 175;
    appendFullwidthWidget(null, ctlr.addTextlabel("Select Joystick")
      .setText("SELECT JOYSTICK:"));
    ScrollableList joylist = ctlr.addScrollableList("Joystick List")
      .setBackgroundColor(color(192))
      .setSize(280, 100)
      .setBarHeight(20)
      .setItemHeight(15)
      .setItems(gJoystick.getJoystickNames(60))
      .plugTo(this, "joystickSelected");
    joylist.getValueLabel().toUpperCase(false);
    joylist.getCaptionLabel().toUpperCase(false);
    joylist.getValueLabel().setFont(textfield.getValueLabel().getFont()); // workaround to get defaultFontForText
    joylist.getCaptionLabel().setFont(textfield.getValueLabel().getFont());
    appendFullwidthWidget("joystickList", joylist);
    setItemsColor(joylist, scrollableListItemColor);

    for (int i=Joystick.MaxButtonNum-1; i>=0; i--) {
      Textlabel label = ctlr.addTextlabel("Button_" + i);
      label.setText("Button " + i)
        .setPosition(x + 10, y + 235 + i * 25)
        .moveTo("Joystick");
      String listLabel = buttonFunctionListName(i);
      ScrollableList buttonFunctionList = ctlr.addScrollableList(listLabel)
        .setBackgroundColor(color(192))
        .setSize(150, 100)
        .setBarHeight(20)
        .setItemHeight(15)
        .setPosition(x + 80, y + 230 + i * 25)
        .setItems(ButtonFunction.getValuesMap())
        .setValue(0)
        .moveTo("Joystick");
      setItemsColor(buttonFunctionList, scrollableListItemColor);
      lockWidget(buttonFunctionList);
      buttonFunctionSelectorsMap.put(listLabel, i);
    }
    joylist.bringToFront();
    int currentJoystickIdx = gJoystick.getCurrentDeviceIndex();
    if (currentJoystickIdx >= 0) joylist.setValue(currentJoystickIdx);

    // Chart tab
    setTab("Chart");
    chart = new ChartCanvas(settings, x + 10, y + 175, w - 20, 300);
    chartJumperXIdx = chart.addSeries("Jumper X");
    chartJumperYIdx = chart.addSeries("Jumper Y");
    chartCameraXIdx = chart.addSeries("Camera X");
    chartCameraYIdx = chart.addSeries("Camera Y");
    chart.pre();
    ctlr.getTab("Chart").addCanvas(chart);
    nextWidgetPosition_y = y + 480;
    Toggle toggle = ctlr.addToggle("Show Velocity");
    toggle.plugTo(this, "showVelocityChartChanged");
    appendHalfwidthWidget("showVelocityChart", toggle);

    // Misc tab
    setTab("Misc");
    nextWidgetPosition_y = y + 175;
    appendHalfwidthWidget("showAfterimage", ctlr.addToggle("Show afterimage"));
    appendHalfwidthWidget("showInputStatus", ctlr.addToggle("Show input status"));

    for (Entry<String, Controller> entry : widgets.entrySet()) {
      Controller widget = entry.getValue();
      if (widget instanceof Toggle) {
        Label caption = widget.getCaptionLabel();
        caption.align(ControlP5.RIGHT_OUTSIDE, ControlP5.CENTER);
        caption.setPadding(5, 0);
      }
    }

    int id = 100;
    for (Entry<String, Controller> entry : widgets.entrySet()) {
      Controller widget = entry.getValue();
      widget.setId(id);
      id2parameter.put(id, entry.getKey());
      id++;
    }

    ctlr.get(ScrollableList.class, "Styles List").bringToFront().setValue(0);
    setControllerValues();

    ctlr.addCallback(new CallbackListener() {
      public void controlEvent(CallbackEvent event) {
        parameterChange(event);
      }
    }
    );

    numIndicator = new NumIndicator();
  }

  public void appendFullwidthWidget(String name, Controller widget) {
    if (halfFilled) {
      halfFilled = false;
      if (lastWidget != null) {
        nextWidgetPosition_y += lastWidget.getHeight() + widgetMargin_y;
      }
    }
    if (widget instanceof Slider) {
      widget.setSize(150, 20);
    }
    widget.setPosition(x + 10, nextWidgetPosition_y);
    widget.moveTo(currentTab);
    if (name != null) widgets.put(name, widget);
    nextWidgetPosition_y += widget.getHeight() + widgetMargin_y;
    lastWidget = widget;
  }

  public void appendHalfwidthWidget(String name, Controller widget) {
    if (widget instanceof Slider) {
      widget.setSize(80, 20);
    }
    if (!halfFilled) {
      widget.setPosition(x + 10, nextWidgetPosition_y);
      halfFilled = true;
    } else {
      widget.setPosition(x + 150, nextWidgetPosition_y);
      halfFilled = false;
      nextWidgetPosition_y += max(widget.getHeight(), lastWidget.getHeight()) + widgetMargin_y;
    }
    widget.moveTo(currentTab);
    if (name != null) widgets.put(name, widget);
    lastWidget = widget;
  }

  public void appendHalfwidthWidget() {
    if (halfFilled) {
      nextWidgetPosition_y += lastWidget.getHeight() + widgetMargin_y;
      halfFilled = false;
    }
  }

  public void statusUpdate(Jumper jumper, Level level, Camera camera) {
    ((Textlabel)widgets.get("JumpingValue")).setText(jumper.jumping?"TRUE":"FALSE");
    ((Textlabel)widgets.get("PropellingValue")).setText(jumper.propelling?"TRUE":"FALSE");
    ((Textlabel)widgets.get("OnObstacleValue")).setText(jumper.onObstacle?"TRUE":"FALSE");
    chart.updateSeries(chartJumperXIdx, jumper.center_x() / level.w);
    chart.updateSeries(chartJumperYIdx, 1.0f - jumper.center_y() / level.h);
    chart.updateSeries(chartCameraXIdx, camera.x / level.w);
    chart.updateSeries(chartCameraYIdx, 1.0f - camera.y / level.h);
    chart.updateChart();
  }

  public void drawStatus(Jumper jumper) {
    drawNumIndicator("JumperXYvalue", String.format("%7.2f,%7.2f", jumper.x, jumper.y));
    drawNumIndicator("JumperVXYvalue", String.format("% 7.2f,% 7.2f", jumper.vx, jumper.vy));
    drawNumIndicator("JumperAXYValue", String.format("% 7.2f,% 7.2f", jumper.vx - jumper.pvx, jumper.vy - jumper.pvy));
    drawNumIndicator("PropellingRemainingFramesValue", String.format("%4d", jumper.propellingRemainingFrames));
    drawButtonsStatus();
  }

  public void drawButtonsStatus() {
    if (ctlr.getTab("Joystick").isActive()) {
      stroke(255);
      for (int i=0; i<gJoystick.buttons.length; i++) {
        if (gJoystick.buttons[i] != null) {
          if (gJoystick.buttons[i].getValue() > 0) {
            fill(255, 255, 0);
          } else {
            fill(64, 64, 0);
          }
        } else {
          fill(128);
        }
        rect (x + 60, y + 234 + i * 25, 10, 10);
      }
    }
  }

  public void drawNumIndicator(String name, String str) {
    int[] pos = indicators.get(name);
    numIndicator.text(str, pos[0], pos[1]);
  }

  public void setControllerValues() {
    for (String name : Settings.booleanVariables) {
      Toggle toggle = (Toggle)widgets.get(name);
      if (toggle != null) {
        try {
          Field f = Class.forName("jumplab$Settings").getDeclaredField(name);
          toggle.setValue((Boolean)(f.get(settings))?1:0);
        }
        catch (ReflectiveOperationException e) {
          System.err.println("Failed to set " + name + ".");
        }
      }
    }
    for (String name : Settings.floatVariables) {
      Slider slider = (Slider)widgets.get(name);
      if (slider != null) {
        try {
          Field f = Class.forName("jumplab$Settings").getDeclaredField(name);
          slider.setValue((float)f.get(settings));
        }
        catch (ReflectiveOperationException e) {
          System.err.println("Failed to set " + name + ".");
        }
      }
    }
    for (String name : Settings.listVariables) {
      ScrollableList slist = (ScrollableList)widgets.get(name);
      if (slist != null) {
        try {
          Field f = Class.forName("jumplab$Settings").getDeclaredField(name);
          String s = (String)f.get(settings);
          int idx = Arrays.asList(gMasao.proportionLabels).indexOf(s);
          if (idx >= 0) {
            slist.setValue(idx);
          } else {
            System.err.println("Invalid character name is found: " + s);
          }
        }
        catch (ReflectiveOperationException e) {
          System.err.println("Failed to set " + name + ".");
        }
      }
    }
  }

  public void parameterChange(CallbackEvent event) {
    Controller widget = event.getController();
    if (widget instanceof Toggle) {
      if (event.getAction() == ControlP5.ACTION_BROADCAST) {
        String name = id2parameter.get(event.getController().getId());
        try {
          Field f = Class.forName("jumplab$Settings").getDeclaredField(name);
          f.set(settings, widget.getValue() > 0);
        }
        catch (ReflectiveOperationException e) {
          System.err.println("Failed to set " + name + ".");
        }
      }
    } else if (widget instanceof Slider) {
      int action = event.getAction();
      if (action == ControlP5.ACTION_MOVE || action == ControlP5.ACTION_RELEASE || action == ControlP5.ACTION_RELEASE_OUTSIDE) {
        String name = id2parameter.get(event.getController().getId());
        try {
          Field f = Class.forName("jumplab$Settings").getDeclaredField(name);
          f.set(settings, widget.getValue());
        }
        catch (ReflectiveOperationException e) {
          System.err.println("Failed to set " + name + ".");
        }
      }
    } else if (widget instanceof Button) {
      if (event.getAction() == ControlP5.ACTION_CLICK) {
        String name = event.getController().getLabel();
        if (name == "Save") {
          String styleName = ctlr.get(Textfield.class, "style name").getText();
          if (styleName.isEmpty()) styleName = "Untitled";
          styles.upsert(styleName, settings);
          styles.save(userSettingsFilename);
          List<String> styleNames = styles.keyList();
          int idx = styleNames.indexOf(styleName);
          if (idx < 0) idx = 0; // if not found, choose the default.
          ScrollableList slist = ctlr.get(ScrollableList.class, "Styles List");
          slist.setItems(styleNames).setValue(idx);
          setItemsColor(slist, scrollableListItemColor);
        }
      }
    } else if (widget instanceof Textfield) {
      String name = event.getController().getLabel();
      if (name == "style name") {
        if (event.getAction() == ControlP5.ACTION_LEAVE) {
          String styleName = ((Textfield)widget).getText();
          changeSaveButtonStatus(styles.isModifiable(styleName));
        }
      }
    } else if (widget instanceof ScrollableList) {
      String name = event.getController().getName();
      if (name.matches(buttonFunctionListName("([0-9]*)"))) {
        if (event.getAction() == ControlP5.ACTION_BROADCAST) {
          assignButtonFunction(name, (ScrollableList)widget, (int)event.getController().getValue());
        }
      }
    }
  }

  public void styleSelected(int value) {
    String key = styles.keyList().get(value);
    Textfield styleName = ctlr.get(Textfield.class, "style name");
    styleName.setValue(key);
    settings.load(styles.get(key));
    setControllerValues();
    changeSaveButtonStatus(styles.isModifiable(key));
  }

  public void styleNameChanged(String str) {
    changeSaveButtonStatus(styles.isModifiable(str));
  }

  public void changeSaveButtonStatus(boolean modifiable) {
    Button saveButton = ctlr.get(Button.class, "Save");
    if (modifiable) {
      unlockWidget(saveButton);
    } else {
      lockWidget(saveButton);
    }
  }

  public void characterSelected(int value) {
    gMasao.setImages(value);
  }

  public void joystickSelected(int value) {
    gJoystick.selectDevice(value);
    gJoystick.saveConfig();
    setButtonAssignmentsValue();
  }

  public void setButtonAssignmentsValue() {
    for (int i=0; i<Joystick.MaxButtonNum; i++) {
      ScrollableList buttonFunctionList = ctlr.get(ScrollableList.class, buttonFunctionListName(i));
      if (gJoystick.buttons[i] != null) {
        unlockWidget(buttonFunctionList);
        int n = gJoystick.buttons[i].buttonFunction.ordinal();
        buttonFunctionList.setValue(n);
      } else {
        buttonFunctionList.setValue(0);
        lockWidget(buttonFunctionList);
      }
    }
  }

  public void assignButtonFunction(String name, ScrollableList widget, int index) {
    gJoystick.assignButtonFunction(buttonFunctionSelectorsMap.get(name), (ButtonFunction) widget.getItem(index).get("value"));
    gJoystick.saveConfig();
  }

  public void showVelocityChartChanged() {
    chart.showVelocityChartChanged();
  }

  public void setTab(String name) {
    currentTab = name;
    halfFilled = false;
  }

  public void setItemsColor(ScrollableList list, CColor col) {
    for (Object itemPtr : list.getItems()) {
      Map<String, Object> item = (Map<String, Object>)itemPtr;
      item.put("color", col);
    }
  }

  public void lockWidget(Controller widget) {
    widget.lock();
    widget.setColorBackground(color(192, 192, 192));
    widget.setColorForeground(color(128, 128, 128));
  }

  public void unlockWidget(Controller widget) {
    widget.unlock();
    CColor mainColor = ControlP5.getColor();
    widget.setColorBackground(mainColor.getBackground());
    widget.setColorForeground(mainColor.getForeground());
  }

  public void buttonFunctionActivated(ButtonFunction func) {
    switch(func) {
    case TOGGLE_TRAIL:
      toggleShowTrail();
      break;
    case TOGGLE_CAMERA:
      toggleShowCameraMarker();
      break;
    case NEXT_STYLE:
      shiftStyle(1);
      break;
    case PREV_STYLE:
      shiftStyle(-1);
      break;
    case PAUSE:
      togglePause();
      break;
    case STEP_FORWARD:
      pressStepForward();
      break;
    default:
      break;
    }
  }

  public void toggleButton(String name) {
    Toggle button = (Toggle)widgets.get(name);
    button.toggle();
  }

  public void toggleShowTrail() {
    toggleButton("showTrail");
  }

  public void toggleShowCameraMarker() {
    toggleButton("showCameraMarker");
  }

  public void togglePause() {
    gPause ^= true;
  }

  public void pressStepForward() {
    gStepForward = true;
  }

  public void shiftStyle(int dir) {
    ScrollableList slist = ctlr.get(ScrollableList.class, "Styles List");
    int slistItemsNum = slist.getItems().size();
    int newValue = ((int)slist.getValue() + dir + slistItemsNum) % slistItemsNum;
    slist.setValue(newValue);
  }

  public String buttonFunctionListName(int i) {
    return "Button " + i + " feature";
  }

  public String buttonFunctionListName(String regexp) {
    return "Button " + regexp + " feature";
  }

  public boolean keyPressed() {
    if (key == 'p') {
      togglePause();
      return true;
    } else if (key == '.') {
      pressStepForward();
    }

    return false;
  }

  public boolean keyReleased() {
    return false;
  }
}




class Joystick {
  static final String joystickConfigVersion = "1.2";
  static final int MaxButtonNum = 12; // 12 would be enough for the most joysticks
  ArrayList<ControlDevice> devices;
  ControlDevice currentDevice;
  ControlSlider slider_x;
  JoyButton[] buttons;
  int prevAxisX = 0;
  boolean[] prevButtonPressed;
  boolean[] buttonPressed;
  static final float stickMargin = 0.2f;
  String configFilename;
  HashMap<String, ButtonFunction[]> buttonAssignmentsMap;

  private class JoyButton {
    ControlButton ctlrButton;
    ButtonFunction buttonFunction;

    JoyButton(ControlButton button) {
      this(button, ButtonFunction.NONE);
    }

    JoyButton(ControlButton button, ButtonFunction func) {
      ctlrButton = button;
      buttonFunction = func;
    }

    public float getValue() {
      return ctlrButton.getValue();
    }
  }

  Joystick(PApplet parent, String configFilename) {
    buttons = new JoyButton[MaxButtonNum];
    prevButtonPressed = new boolean[ButtonFunction.size];
    buttonPressed = new boolean[ButtonFunction.size];
    devices = new ArrayList<ControlDevice>();

    ControlIO ctrlio = ControlIO.getInstance(parent);
    for (ControlDevice dev : ctrlio.getDevices()) {
      if ((dev.getTypeID() & (GCP.STICK | GCP.GAMEPAD)) != 0) {
        devices.add(dev);
      }
    }
    this.configFilename = configFilename;
    loadConfig();
  }

  public boolean loadConfig() {
    JSONObject configJson;
    buttonAssignmentsMap = new HashMap<String, ButtonFunction[]>();

    /* Processing 3.5.4 does not return null if the file is not found
     and throws NullPointerException. As a workaround, we check the
     file exists or not before open the file. */
    BufferedReader reader;
    reader = createReader(configFilename);
    if (reader == null) {
      return false;
    }
    println("loading '" + configFilename + "'...");

    try {
      configJson = new JSONObject(reader);
      reader.close();
    }
    catch (IOException e) {
      e.printStackTrace();
      return false;
    }
    catch (RuntimeException e) {
      System.err.println("Failed to load '" + configFilename + "'.");
      e.printStackTrace();
      return false;
    }

    loadJoystickConfigs(configJson);

    String lastSelectedDeviceName = configJson.getString("lastSelected", null);
    if (lastSelectedDeviceName != null) {
      for (ControlDevice dev : devices) {
        if (dev.getName().equals(lastSelectedDeviceName)) {
          selectDevice(dev);
          break;
        }
      }
    }

    return true;
  }

  public void loadJoystickConfigs(JSONObject jsonObject) {
    JSONArray configsJson = jsonObject.getJSONArray("configs");
    if (configsJson == null) return;

    for (int i=0; i<configsJson.size(); i++) {
      JSONObject configJson = configsJson.getJSONObject(i);
      JSONArray buttonAssignmentsJson = configJson.getJSONArray("buttons");
      ArrayList<ButtonFunction> assignments = new ArrayList<ButtonFunction>();
      for (int j=0; j<buttonAssignmentsJson.size(); j++) {
        try {
          assignments.add(ButtonFunction.valueOf(buttonAssignmentsJson.getString(j)));
        }
        catch (IllegalArgumentException e) {
          System.err.println("Warning: unsupported button function was found ('" + buttonAssignmentsJson.getString(j) + "')");
        }
      }
      buttonAssignmentsMap.put(configJson.getString("name"), assignments.toArray(new ButtonFunction[0]));
    }
  }

  public void saveConfig() {
    JSONObject configJson = new JSONObject();
    configJson.setString("lastSelected", currentDevice.getName());
    configJson.setString("version", joystickConfigVersion);
    JSONArray buttonAssignmentsJson = dumpButtonAssignmentsToJson();
    configJson.setJSONArray("configs", buttonAssignmentsJson);
    saveJSONObject(configJson, "data/" + configFilename);
  }

  public JSONArray dumpButtonAssignmentsToJson() {
    JSONArray configsJson = new JSONArray();
    for (Entry<String, ButtonFunction[]> entry : buttonAssignmentsMap.entrySet()) {
      JSONObject configJson = new JSONObject();
      JSONArray buttonAssignmentsJson = new JSONArray();
      ButtonFunction[] functions = entry.getValue();
      for (ButtonFunction function : functions) {
        buttonAssignmentsJson.append(function.toString());
      }
      configJson.setString("name", entry.getKey());
      configJson.setJSONArray("buttons", buttonAssignmentsJson);
      configsJson.append(configJson);
    }
    return configsJson;
  }

  public List<String> getJoystickNames(int maxLength) {
    ArrayList<String> list = new ArrayList<String>();
    for (ControlDevice dev : devices) {
      String name = dev.getName();
      if (name.length() > maxLength) {
        name = name.substring(0, maxLength - 3) + "...";
      }
      list.add(name);
    }
    return list;
  }

  public void update(int[] res) {
    if (currentDevice == null) return;
    currentDevice.update();
    axesUpdate(res);
    buttonsUpdate(res);
  }

  public void updateDuringPause() {
    if (currentDevice == null) return;
    currentDevice.update();
    buttonsUpdateWhilePausing();
  }

  public void axesUpdate(int[] res) {
    if (gPause && !gStepForward) return;
    int axisX = 0;
    float axisXRawValue = slider_x.getValue();

    if (axisXRawValue < -stickMargin) {
      axisX = -1;
    } else if (axisXRawValue > stickMargin) {
      axisX = 1;
    }
    if (prevAxisX == axisX) {
      res[0] = 0;
      res[1] = 0;
    } else {
      if (prevAxisX < 0) {
        res[0] = -1;
      } else if (prevAxisX > 0) {
        res[1] = -1;
      }
      if (axisX < 0) {
        res[0] = 1;
      } else if (axisX > 0) {
        res[1] = 1;
      }
    }
    prevAxisX = axisX;
  }

  public void buttonsUpdate(int[] res) {
    if (gPause && !gStepForward) {
      buttonsUpdateWhilePausing();
      return;
    }

    for (int i=0; i<ButtonFunction.size; i++) {
      prevButtonPressed[i] = buttonPressed[i];
      buttonPressed[i] = false;
    }
    for (int i=0; i<buttons.length; i++) {
      if (buttons[i] != null) {
        ButtonFunction func = buttons[i].buttonFunction;
        if (func != ButtonFunction.NONE) {
          if (buttons[i].getValue() > 0) {
            buttonPressed[func.ordinal()] |= true;
          }
        }
      }
    }
    for (int i=ButtonFunction.JUMP.ordinal(); i<=ButtonFunction.DASH.ordinal(); i++) {
      int s;
      if (prevButtonPressed[i]) {
        s = buttonPressed[i]?0:-1;
      } else {
        s = buttonPressed[i]?1:0;
      }
      res[i - ButtonFunction.JUMP.ordinal() + 2] = s;
    }
    for (int i=ButtonFunction.TOGGLE_TRAIL.ordinal(); i<ButtonFunction.size; i++) {
      if (buttonPressed[i] && !prevButtonPressed[i]) {
        gConsole.buttonFunctionActivated(ButtonFunction.getEnumByOrdinal(i));
      }
    }
  }

  public void buttonsUpdateWhilePausing() {
    for (int i=0; i<ButtonFunction.size; i++) {
      if (ButtonFunction.isEnabledWhilePausing(i)) {
        prevButtonPressed[i] = buttonPressed[i];
        buttonPressed[i] = false;
      }
    }
    for (int i=0; i<buttons.length; i++) {
      if (buttons[i] != null) {
        ButtonFunction func = buttons[i].buttonFunction;
        if (func.isEnabledWhilePausing()) {
          if (buttons[i].getValue() > 0) {
            buttonPressed[func.ordinal()] |= true;
          }
        }
      }
    }
    for (int i=ButtonFunction.TOGGLE_TRAIL.ordinal(); i<ButtonFunction.size; i++) {
      if (ButtonFunction.isEnabledWhilePausing(i) && buttonPressed[i] && !prevButtonPressed[i]) {
        gConsole.buttonFunctionActivated(ButtonFunction.getEnumByOrdinal(i));
      }
    }
  }

  public ControlDevice selectDevice(int idx) {
    return selectDevice(devices.get(idx));
  }

  public ControlDevice selectDevice(ControlDevice device) {
    if (currentDevice != null) {
      currentDevice.close();
    }
    slider_x = null;
    Arrays.fill(buttons, null);
    currentDevice = device;
    currentDevice.open();
    int sliders = currentDevice.getNumberOfSliders();
    for (int i=0; i<sliders; i++) {
      ControlSlider s = currentDevice.getSlider(i);
      if (s.getName().matches("[xX].*")) { // workaround needed because GCP hides the component's identifier.
        slider_x = s;
        break;
      }
    }
    int buttonsNum = min(MaxButtonNum, currentDevice.getNumberOfButtons());
    for (int i=0; i<buttonsNum; i++) {
      buttons[i] = new JoyButton(currentDevice.getButton(i));
    }
    restoreConfig(currentDevice);

    return currentDevice;
  }

  public void restoreConfig(ControlDevice device) {
    ButtonFunction[] buttonAssignments = buttonAssignmentsMap.get(device.getName());
    if (buttonAssignments == null) {
      int num = min(MaxButtonNum, device.getNumberOfButtons());
      buttonAssignments = new ButtonFunction[num];
      for (int i=0; i<num; i++) {
        if (i < 4) {
          buttons[i].buttonFunction = ButtonFunction.JUMP;
          buttonAssignments[i] = ButtonFunction.JUMP;
        } else {
          buttons[i].buttonFunction = ButtonFunction.NONE;
          buttonAssignments[i] = ButtonFunction.NONE;
        }
      }
      buttonAssignmentsMap.put(device.getName(), buttonAssignments);
    } else {
      int num = min(MaxButtonNum, device.getNumberOfButtons(), buttonAssignments.length);
      for (int i=0; i<num; i++) {
        buttons[i].buttonFunction = buttonAssignments[i];
      }
    }
  }

  public int getCurrentDeviceIndex() {
    return devices.indexOf(currentDevice);
  }

  public void assignButtonFunction(int buttonNum, ButtonFunction func) {
    if (buttons[buttonNum] != null) {
      buttons[buttonNum].buttonFunction = func;
    }
    ButtonFunction[] buttonAssignments = buttonAssignmentsMap.get(currentDevice.getName());
    if (buttonNum < buttonAssignments.length) {
      buttonAssignments[buttonNum] = func;
    }
  }
}


class Jumper {
  float x, y;
  float px, py;
  float vx, vy;
  float pvx, pvy;
  float ay;
  float maxVx;
  int dir;
  int lastDir;
  int jumpDir;
  boolean jumping;
  boolean dashing;
  boolean standing;
  boolean propelling;
  boolean onObstacle;
  int jumpMotion;
  int jumpMotionMax;
  int propellingRemainingFrames;
  float pattern;
  int runningMotionMax;
  Settings settings;
  Level level;
  int[] keyInput;
  int[] joyInput;
  boolean[] inputStatus;
  static final int inputTypes = 4; // left, right, jump, dash

  Vector<Vector<PImage>> stock_images_running_l;
  Vector<Vector<PImage>> stock_images_running_r;
  Vector<Vector<PImage>> stock_images_jumping_l;
  Vector<Vector<PImage>> stock_images_jumping_r;
  final String[] proportionLabels = {"1h", "3h", "8h"};

  int character;
  Vector<PImage> images_running_r;
  Vector<PImage> images_running_l;
  Vector<PImage> images_jumping_r;
  Vector<PImage> images_jumping_l;

  static final int w = 24;
  static final int h = 48;

  Jumper(Settings settings, Level level) {
    this.settings = settings;
    this.level = level;

    x = (float)level.sx;
    y = (float)level.sy;
    px = x;
    py = y;
    vx = vy = 0;
    pvx = pvy = 0;
    jumping = false;
    propelling = false;
    onObstacle = false;
    jumpMotion = 0;
    dir = 0;
    lastDir = 1;
    jumpDir = 0;
    pattern = 0;
    ay = settings.gravityFalling;
    keyInput = new int[inputTypes];
    joyInput = new int[inputTypes];
    inputStatus = new boolean[inputTypes];

    initializeImages();
    setImages(1);
  }

  public void initializeImages() {
    stock_images_running_r = new Vector<Vector<PImage>>();
    stock_images_running_l = new Vector<Vector<PImage>>();
    stock_images_jumping_r = new Vector<Vector<PImage>>();
    stock_images_jumping_l = new Vector<Vector<PImage>>();

    for (String label : proportionLabels) {
      images_running_r = new Vector<PImage>();
      images_running_l = new Vector<PImage>();
      images_jumping_r = new Vector<PImage>();
      images_jumping_l = new Vector<PImage>();

      runningMotionMax = 8;
      for (int i=0; i<runningMotionMax+1; i++) {
        String filename = String.format("%s/stickman-%d.png", label, i);
        images_running_r.add(loadImage(filename));
        images_running_l.add(hflipImage(images_running_r.get(i)));
      }
      jumpMotionMax = 4;
      for (int i=0; i<jumpMotionMax; i++) {
        String filename = String.format("%s/stickman-jump-%d.png", label, jumpMotionMax - i);
        images_jumping_r.add(loadImage(filename));
        images_jumping_l.add(hflipImage(images_jumping_r.get(i)));
      }

      stock_images_running_r.add(images_running_r);
      stock_images_running_l.add(images_running_l);
      stock_images_jumping_r.add(images_jumping_r);
      stock_images_jumping_l.add(images_jumping_l);
    }
  }

  public void setImages(int num) {
    settings.selectedCharacter = proportionLabels[num];
    images_running_r = stock_images_running_r.get(num);
    images_running_l = stock_images_running_l.get(num);
    images_jumping_r = stock_images_jumping_r.get(num);
    images_jumping_l = stock_images_jumping_l.get(num);
  }

  public float center_x() {
    return x + Jumper.w / 2;
  }

  public float center_y() {
    return y + Jumper.h / 2;
  }

  public void update() {
    boolean hDL, hDR, hUL, hUR;

    maxVx = dashing?settings.maxVxDashing:settings.maxVx;

    pvx = vx;
    pvy = vy;
    px = x;
    py = y;

    control(keyInput);
    joystickUpdate();
    velocityXUpdate();
    velocityYUpdate();
    animationUpdate();

    x += vx;
    y += vy;

    hDL = hitDL(0);
    hDR = hitDR(0);

    if (!onObstacle && (hDL || hDR)) {
      float ny = y - level.obstaclePenaltyD(y);
      if (vy > 0) {
        if ((hDL && !hitUL()) || (hDR && !hitUR())) {
          if (py <= ny) {
            y = ny;
            jumping = false;
            onObstacle = true;
            vy = 0;
          } else {
            float dy = level.obstaclePenaltyD(y);
            if (dy < settings.vCollisionTolerance) {
              y -= dy;
            }
          }
        }
      }
    } else if (!hitBL() && !hitBR()) {
      onObstacle = false;
      if (!jumping && settings.stopAndFall) {
        vx = 0;
      }
    }

    if (hitUL() && hitUR()) {
      y += level.obstaclePenaltyU(y);
      vy = 0;
      propelling = false;
    }

    hUL = hitUL();
    hDL = hitDL(0);
    if (hUL || hDL) {
      float penalty = level.obstaclePenaltyL(x);
      if (hUL && !hDL) {
        float ny = y + level.obstaclePenaltyU(y);
        if (vy < 0 && penalty >= settings.collisionTolerance && ny <= py) {
          y += level.obstaclePenaltyU(y);
          vy = 0;
          propelling = false;
        } else {
          x += penalty ;
          if (vx < 0) vx = 0;
        }
      } else {
        x += penalty;
        if (vx < 0) vx = 0;
      }
    }

    hUR = hitUR();
    hDR = hitDR(0);
    if (hUR || hDR) {
      float penalty = level.obstaclePenaltyR(x + w);
      if (hUR && !hDR) {
        float ny = y + level.obstaclePenaltyU(y);
        if (vy < 0 && penalty >= settings.collisionTolerance && ny <= py) {
          y += level.obstaclePenaltyU(y);
          vy = 0;
          propelling = false;
        } else {
          x -= penalty;
          if (vx > 0) vx = 0;
        }
      } else {
        x -= penalty;
        if (vx > 0) vx = 0;
      }
    }
  }

  public void velocityXUpdate() {
    float ax;
    if (jumping || (!onObstacle && !settings.allowAerialWalk)) {
      ax = settings.axJumping;
    } else if (!jumping && !onObstacle && settings.stopAndFall) {
      ax = 0;
    } else if (Math.signum(vx) != dir) {
      ax = settings.axBrake;
    } else {
      if (dashing) {
        ax = settings.axNormalDashing;
      } else {
        ax = settings.axNormal;
      }
    }
    if (dir != 0) {
      vx += ax * dir;
      if (settings.vxAdjustmentAtTakeoff > 0 && jumping) {
        vx = constrain(vx, -maxVx * (1 + settings.vxAdjustmentAtTakeoff), maxVx * (1 + settings.vxAdjustmentAtTakeoff));
      } else {
        vx = constrain(vx, -maxVx, maxVx);
      }
    } else {
      vx -= ax * Math.signum(vx);
      if (abs(vx) < ax) {
        vx = 0;
      }
    }
  }

  public void velocityYUpdate() {
    if (jumping && jumpMotion > 0) {
      jumpMotion--;
      if (jumpMotion == 0) {
        vy = -settings.jumpVelocity - (settings.jumpVelocityBonus * abs(vx));
        if (settings.vxAdjustmentAtTakeoff < 0) {
          vx += vx * settings.vxAdjustmentAtTakeoff;
        } else if (settings.vxAdjustmentAtTakeoff > 0) {
          vx += vx * settings.vxAdjustmentAtTakeoff;
        }
        onObstacle = false;
        if (settings.maxPropellingFrames > 0) {
          propelling = true;
          propellingRemainingFrames = round(settings.maxPropellingFrames);
        }
      }
    }
    if (!onObstacle) {
      if (!settings.allowAerialJump && jumping && jumpMotion > 0) {
        jumping = false;
        jumpMotion = 0;
        propelling = false;
      }
      if (vy > 0) {
        ay = settings.gravityFalling;
      }
      if (!propelling) {
        vy += ay;
        if (vy > settings.maxVy) {
          vy = settings.maxVy;
        }
      }
      if (settings.maxPropellingFrames > 0) {
        if (propelling) {
          propellingRemainingFrames--;
          if (propellingRemainingFrames <= 0) {
            propelling = false;
          }
        }
      }
      if (settings.allowWallSlide) {
        if (lastDir < 0 && hitDL(1) && inputStatus[0] || lastDir > 0 && hitDR(1) && inputStatus[1]) {
          if (vy >= settings.maxVy * 0.2f) vy = settings.maxVy * 0.2f;
        }
      }
    }
  }

  public void animationUpdate() {
    if (vx == 0) {
      int p = (int)pattern;
      if (p % (runningMotionMax / 2) > 0) {
        pattern += 0.5f;
      } else {
        p = 0;
        standing = true;
      }
    } else {
      pattern += abs(vx/12);
      standing = false;
    }
    if (pattern >= runningMotionMax) {
      pattern -= runningMotionMax;
    }
  }

  public boolean hitUL() {
    return level.isThereObstacle((int)x, (int)y);
  }
  public boolean hitUR() {
    return level.isThereObstacle(ceil(x + w - 1), (int)y);
  }
  public boolean hitDL(int d) {
    return level.isThereObstacle((int)x - d, ceil(y + h - 1));
  }
  public boolean hitDR(int d) {
    return level.isThereObstacle(ceil(x + w - 1) + d, ceil(y + h - 1));
  }
  public boolean hitBL() {
    return level.isThereObstacle((int)x, ceil(y + h));
  }
  public boolean hitBR() {
    return level.isThereObstacle(ceil(x + w - 1), ceil(y + h));
  }

  public void move(int d) {
    dir = d;
    if (dir != 0) {
      lastDir = dir;
    }
  }

  public void jump() {
    if (!jumping) {
      if (onObstacle) {
        jumping = true;
        jumpDir = lastDir;
        jumpMotion = round(settings.jumpAnticipationFrames) + 1;
        ay = settings.gravity;
      }
    }
    if (!onObstacle && settings.allowWallJump) {
      if (lastDir < 0 && hitDL(1) && inputStatus[0] || lastDir > 0 && hitDR(1) && inputStatus[1]) {
        jumping = true;
        jumpDir = -lastDir;
        jumpMotion = 1;
        ay = settings.gravity;
        vx = -maxVx * lastDir * settings.wallJumpSpeedRatio;
        vy = 0;
      }
    }
  }

  public void jumpCanceled() {
    if (jumping) {
      ay = settings.gravityFalling;
      if (vy < 0) vy *= settings.verticalSpeedSustainLevel;
      propelling = false;
    }
  }

  public void dashStart() {
    dashing = true;
  }

  public void dashEnd() {
    dashing = false;
  }

  public void draw(PImage img, float dx, float dy) {
    if (img != null) {
      image(img, dx - (img.width - 24) / 2, dy);
    }
  }

  public PImage draw(float cx, float cy) {
    PImage img;

    if (jumping) {
      int p = (jumpMotion == 0)?0: (int)(jumpMotion * (jumpMotionMax - 1) / (settings.jumpAnticipationFrames + 1) + 1);
      int dir = settings.allowAerialTurn?lastDir:jumpDir;
      if (dir < 0) {
        img = images_jumping_l.get(p);
      } else {
        img = images_jumping_r.get(p);
      }
    } else if (standing) {
      img = (lastDir < 0)?images_running_l.get(0):images_running_r.get(0);
    } else {
      int p = (int)pattern;
      if (lastDir < 0) {
        img = images_running_l.get(p + 1);
      } else {
        img = images_running_r.get(p + 1);
      }
    }
    this.draw(img, x - cx, y - cy);

    return img;
  }

  public boolean keyPressed() {
    if (key == CODED) {
      if (keyCode == LEFT) {
        keyInput[0] = 1;
        return true;
      } else if (keyCode == RIGHT) {
        keyInput[1] = 1;
        return true;
      } else if (keyCode == SHIFT) {
        keyInput[3] = 1;
      }
    } else if (key == ' ') {
      keyInput[2] = 1;
      return true;
    }
    return false;
  }

  public boolean keyReleased() {
    if (key == CODED) {
      if (keyCode == LEFT) {
        keyInput[0] = -1;
        return true;
      } else if (keyCode == RIGHT) {
        keyInput[1] = -1;
        return true;
      } else if (keyCode == SHIFT) {
        keyInput[3] = -1;
      }
    } else if (key == ' ') {
      keyInput[2] = -1;
      return true;
    }
    return false;
  }

  public void control(int[] input) {
    if (input[0] > 0) {
      move(-1);
    } else if (input[1] > 0) {
      move(1);
    } else if (input[0] < 0 && dir == -1 || input[1] < 0 && dir == 1) {
      move(0);
    }
    if (input[2] > 0) {
      jump();
    } else if (input[2] < 0) {
      jumpCanceled();
    }
    if (input[3] > 0) {
      dashStart();
    } else if (input[3] < 0) {
      dashEnd();
    }
    for (int i=0; i<inputTypes; i++) {
      if (input[i] > 0) inputStatus[i] = true;
      if (input[i] < 0) inputStatus[i] = false;
      input[i] = 0;
    }
  }

  public void joystickUpdate() {
    gJoystick.update(joyInput);
    control(joyInput);
  }

  public PImage hflipImage(PImage srcImg) {
    PImage destImg = createImage(srcImg.width, srcImg.height, ARGB);
    int x, y;

    srcImg.loadPixels();
    destImg.loadPixels();
    for (y=0; y<srcImg.height; y++) {
      for (x=0; x<srcImg.width; x++) {
        destImg.pixels[srcImg.width - 1 - x + y * srcImg.width] =
          srcImg.pixels[x + y * srcImg.width];
      }
    }
    destImg.updatePixels();
    return destImg;
  }
}
class Level {
  int w, h; // level size in pixel
  int cw, ch; // level size in character
  int bw = 48; // block size
  int bh = 48;
  int sx, sy; // initial position of the player character
  int[][] map;
  PImage bgImg;
  PImage tileImg;

  Level(String mapFile, String blockImgFile, String bgFile) {
    sx = sy = 1;
    Table mapdata = loadTable(mapFile);
    cw = mapdata.getColumnCount();
    ch = mapdata.getRowCount();
    map = new int[ch][cw];
    for (int y=0; y<ch; y++) {
      TableRow row = mapdata.getRow(y);
      for (int x=0; x<cw; x++) {
        int v = row.getInt(x);
        if (v == 9) {
          sx = x;
          sy = y;
        } else {
          map[y][x] = v;
        }
      }
    }

    tileImg = loadImage(blockImgFile);
    bgImg = loadImage(bgFile);

    w = bw * cw;
    h = bh * ch;

    sx = sx * bw;
    sy = sy * bh;
  }

  public PImage createLevelImage() {
    PGraphics levelGfx = createGraphics(w, h);
    levelGfx.beginDraw();
    levelGfx.background(0, 0);
    for (int y=0; y<ch; y++) {
      for (int x=0; x<cw; x++) {
        int chip = map[y][x];
        if (chip > 0 && chip < 9) {
          levelGfx.copy(tileImg, (chip - 1) * bw, 0, bw, bh, x * bw, y * bh, bw, bh);
        }
      }
    }
    levelGfx.endDraw();
    return levelGfx.get();
  }

  public int getChip(int x, int y) {
    int cx = floor((float)x / bw);
    int cy = floor((float)y / bh);
    if (cx < 0 || cy < 0 || cx >= cw || cy >= ch) return 1;
    return map[cy][cx];
  }

  public boolean isThereObstacle(int x, int y) {
    return getChip(x, y) > 0;
  }

  public float obstaclePenaltyL(float x) {
    int cx = ceil(x / bw);
    return cx * bw - x;
  }

  public float obstaclePenaltyR(float x) {
    int cx = (int)x / bw;
    return x - cx * bw;
  }

  public float obstaclePenaltyU(float y) {
    int cy = ceil(y / bh);
    return cy * bw - y;
  }

  public float obstaclePenaltyD(float y) {
    int cy = (int)y / bh;
    return y - cy * bw;
  }
}
class NumIndicator {
  // This class would not be needed if ControlP5 had provided any monospace font...
  static final int glyph_w = 6;
  static final int glyph_h = 8;

  PImage[] glyphs;
  NumIndicator() {
    glyphs = new PImage[15];
    PImage img = loadImage("numglyphs.png");
    img.loadPixels();
    for (int i=0; i<15; i++) {
      glyphs[i] = createImage(glyph_w, glyph_h, RGB);
      glyphs[i].loadPixels();
      for (int y=0; y<glyph_h; y++) {
        for (int x=0; x<glyph_w; x++) {
          glyphs[i].pixels[y * glyph_w + x] = img.pixels[(y * 15 + i) * glyph_w + x];
        }
      }
      glyphs[i].updatePixels();
    }
  }

  public void text(String str, int x, int y) {
    for (char ch : str.toCharArray()) {
      int idx = 0;
      if (ch >= ',' && ch <= '9') {
        idx = ch - ',' + 1;
      }
      fill(0);
      image(glyphs[idx], x, y);
      x += glyph_w;
    }
  }
}





static class Settings {
  static final String styleVersion = "1.3.0";
  // Jump parameters
  boolean showTrail = false; // Show the trail or not.
  boolean allowAerialJump = true; // Allow aerial jump or not.
  boolean allowAerialWalk = true; // Allow aerial walk or not.
  boolean allowAerialTurn = false; // Allow aerial turn or not. (Only affects the visual appearance)
  boolean stopAndFall = false; // The horizontal motion is halted when the jumper goes off the foothold.
  boolean allowWallJump = false; // Allow wall jump or not.
  boolean allowWallSlide = false; // Allow wall slide or not.
  float maxVx = 8; // Maximum horizontal velocity of the jumper.
  float maxVy = 30; // Maximum vertical velocity of the jumper. Limits only the falling motion.
  float jumpVelocity = 13; // Initial vertical velocity of a jump motion.
  float jumpVelocityBonus = 0;// The faster run gives an initial jump velocity bonus that allows a higher jump.
  float jumpAnticipationFrames = 1; // Duration of the anticipation of jump motion in frames.
  float vxAdjustmentAtTakeoff = 0.0f; // Horizontal velocity adjustment at the takeoff.
  float maxPropellingFrames = 0; // Maximum duration of propelled jump.
  float gravity = 0.5f; // gravity when rising.
  float gravityFalling = 1.2f; // gravity when falling.
  float verticalSpeedSustainLevel = 1.0f; // Sustain level of the vertical speed when the button released.
  float axNormal = 0.2f; // Horizontal acceleration in normal state.
  float axBrake = 1.0f; // Horizontal acceleration when braking.
  float axJumping = 0.1f; // Horizontal acceleration when jumping.
  float collisionTolerance = 8; // Tolerance to automatically avoid blocks when jumping (in pixels).
  float vCollisionTolerance = 0; // Vertical tolerance to automatically avoid block when falling (in pixels).
  float wallJumpSpeedRatio = 1.0f; // The velocity raito of the jumping speed of the wall jump to the maxVx.
  float maxVxDashing = 12; // Maximum horizontal velocity of the jumper while dashing.
  float axNormalDashing = 0.4f; // Horizontal acceleration in normal state while dashing.

  // Camera parameters
  boolean showCameraMarker = false; // Show the center marker or not.
  boolean cameraEasing_x = false; // Ease the horizon camera motion or not.
  boolean cameraEasing_y = true; // Ease the vertical camera motion or not.
  boolean forwardFocus = false; //  Shift the focus to the front of the jumper to enable wide forward view.
  boolean projectedFocus = false; // Shift the focus to the projeced position of the jumber, which means running faster moves the focal point farther.
  boolean platformSnapping = false; // Halt the vertical camera motion when the jumper is jumping.
  float cameraEasingNormal_x = 0.1f; // Smoothness of the horizontal camera motion in normal state.
  float cameraEasingNormal_y = 0.1f; // Smoothness of the vertical camera motion in normal state.
  float cameraEasingGrounding_y = 0.2f; // Smoothness of the camera motion when the jumper grounded.
  float cameraWindow_h = 0; // Height of the camera window.
  float cameraWindow_w = 0; // Width of the camera window.
  float focusDistance = 100; // Distance to the focal point.
  float focusingSpeed = 5; // Velocity of the focal point movement.
  float bgScrollRatio = 0.5f; // Determine the scroll speed of the BG layer.

  // Character parameters
  String selectedCharacter = "3H"; // Name of the selected character.

  // Misc. parameters
  boolean showVelocityChart = false; // Show velocity chart in the chart canvas.
  boolean showAfterimage = false; // Show afterimage instead of red dots when 'showTrail' is true.
  boolean showInputStatus = false; // Show the input status.

  static ArrayList<String> booleanVariables;
  static ArrayList<String> floatVariables;
  static ArrayList<String> listVariables = new ArrayList<String>(Arrays.asList("selectedCharacter"));
  static List<String> ignoredVariables = Arrays.asList("showTrail", "showAfterimage", "showCameraMarker", "showVelocityChart", "showInputStatus");

  static {
    booleanVariables = new ArrayList<String>();
    floatVariables = new ArrayList<String>();
    Field[] allVariables = Settings.class.getDeclaredFields();
    for (Field f : allVariables) {
      if (!Modifier.isStatic(f.getModifiers())) {
        if (f.getType() == Boolean.TYPE) {
          booleanVariables.add(f.getName());
        } else if (f.getType() == Float.TYPE) {
          floatVariables.add(f.getName());
        }
      }
    }
  }

  Settings() {
    // Nothing to do at this momemnt. Just a placeholder.
  }

  public void load(Style style) {
    for (String key : style.keySet()) {
      try {
        Field variable = this.getClass().getDeclaredField(key);
        Object value = style.get(key);
        variable.set(this, value);
      }
      catch (ReflectiveOperationException e) {
        System.err.println("Failed to get " + key + " of " + style.getName() + ".");
      }
    }
  }

  public HashMap<String, Object> toHashMap() {
    HashMap<String, Object> map = new HashMap<String, Object>();
    for (String variableName : booleanVariables) {
      if (ignoredVariables.contains(variableName)) continue;
      try {
        Field variable = this.getClass().getDeclaredField(variableName);
        map.put(variableName, variable.getBoolean(this));
      }
      catch (ReflectiveOperationException e) {
        System.err.println("Failed to get " + variableName + " from the current settings.");
      }
    }
    for (String variableName : floatVariables) {
      if (ignoredVariables.contains(variableName)) continue;
      try {
        Field variable = this.getClass().getDeclaredField(variableName);
        map.put(variableName, variable.getFloat(this));
      }
      catch (ReflectiveOperationException e) {
        System.err.println("Failed to get " + variableName + " from the current settings.");
      }
    }
    for (String variableName : listVariables) {
      if (ignoredVariables.contains(variableName)) continue;
      try {
        Field variable = this.getClass().getDeclaredField(variableName);
        map.put(variableName, (String)variable.get(this));
      }
      catch (ReflectiveOperationException e) {
        System.err.println("Failed to get " + variableName + " from the current settings.");
      }
    }
    return map;
  }
}



class Style {
  private String name;
  private boolean modifiable;
  private ComparableVersion version;
  private HashMap<String, Object> data;
  private OffsetDateTime lastModified;

  Style(JSONObject json) {
    name = json.getString("name");
    modifiable = json.getBoolean("modifiable", true);
    version = new ComparableVersion(json.getString("version", "1.0"));
    String lastModifiedStr = json.getString("last modified", "2020-08-01T00:00:00.000+09:00");
    lastModified = OffsetDateTime.parse(lastModifiedStr);
    data = new HashMap<String, Object>();
    JSONObject jsonData = json.getJSONObject("data");
    if (jsonData == null) {
      jsonData = new JSONObject();
    }

    if (version.compareTo("1.2") < 0) {
      boolean parallaxScrolling = jsonData.getBoolean("parallaxScrolling", true);
      if (parallaxScrolling) {
        jsonData.setFloat("bgScrollRatio", 0.5f);
        jsonData.put("parallaxScrolling", null);
      } else {
        jsonData.setFloat("bgScrollRatio", 1.0f);
        jsonData.put("parallaxScrolling", null);
      }
    } else if (version.compareTo("1.2.1") < 0) {
      jsonData.setBoolean("allowWallJump", false);
    } else if (version.compareTo("1.2.2") < 0) {
      jsonData.setString("selectedCharacter", "3h");
      jsonData.setFloat("vCollisionTolerance", 0);
    }

    for (Object key : jsonData.keys()) {
      String variableName = (String)key;
      if (Settings.booleanVariables.contains(variableName)) {
        data.put(variableName, jsonData.getBoolean(variableName));
      } else if (Settings.floatVariables.contains(variableName)) {
        data.put(variableName, jsonData.getFloat(variableName));
      } else if (Settings.listVariables.contains(variableName)) {
        data.put(variableName, jsonData.getString(variableName));
      }
    }
  }

  Style(String name) {
    this.name = name;
    modifiable = true;
  }

  public String getName() {
    return name;
  }

  public boolean isModifiable() {
    return modifiable;
  }

  public Set<String> keySet() {
    return data.keySet();
  }

  public Object get(String key) {
    return data.get(key);
  }

  public void updateLastModified() {
    lastModified = OffsetDateTime.now();
  }

  public boolean update(Settings settings) {
    if (modifiable) {
      data = settings.toHashMap();
      updateLastModified();
      version = new ComparableVersion(Settings.styleVersion);
      return true;
    }
    return false;
  }

  public JSONObject toJson() {
    JSONObject json = new JSONObject();
    json.setString("name", name);
    json.setBoolean("modifiable", modifiable);
    json.setString("last modified", lastModified.toString());
    json.setString("version", version.toString());

    JSONObject jsonData = new JSONObject();
    for (Entry<String, Object> entry : data.entrySet()) {
      jsonData.put(entry.getKey(), entry.getValue());
    }
    json.setJSONObject("data", jsonData);

    return json;
  }
}



class StyleManager {
  HashMap <String, Style> styles;

  StyleManager() {
    styles = new HashMap<String, Style>();
  }

  public boolean loadDefaultSettings(String filename) {
    return load(filename, true);
  }

  public boolean loadUserSettings(String filename) {
    return load(filename, false);
  }

  private boolean load(String filename, boolean isDefaultSettings) {
    JSONArray stylesJson;

    /* Processing 3.5.4 does not return null if the file is not found
     and throws NullPointerException. As a workaround, we check the
     file exists or not before open the file. */
    BufferedReader reader;
    reader = createReader(filename);
    if (reader == null) {
      return false;
    }
    println("loading '" + filename + "'...");

    try {
      stylesJson = new JSONArray(reader);
      reader.close();
    }
    catch (IOException e) {
      e.printStackTrace();
      return false;
    }
    catch (RuntimeException e) {
      System.err.println("Failed to load '" + filename + "'.");
      e.printStackTrace();
      return false;
    }

    for (int i=0; i<stylesJson.size(); i++) {
      JSONObject styleJson = stylesJson.getJSONObject(i);
      Style style = new Style(styleJson);
      if (isDefaultSettings || style.isModifiable()) {
        styles.put(style.name, style);
      }
    }

    return true;
  }

  public Style get(String name) {
    return styles.get(name);
  }

  public boolean upsert(String name, Settings settings) {
    Style style = get(name);
    if (style == null) {
      style = new Style(name);
      styles.put(name, style);
    }
    return style.update(settings);
  }

  public void save(String filename) {
    JSONArray stylesJson = new JSONArray();
    for (Style style : styles.values()) {
      if (style.isModifiable() && !style.getName().equals("*default*")) {
        stylesJson.append(style.toJson());
      }
    }
    saveJSONArray(stylesJson, "data/" + filename);
  }

  public List<String> keyList() {
    ArrayList<String> keys = new ArrayList<String>(styles.keySet());
    Collections.sort(keys, new Comparator<String>() {
      @Override public int compare(String s1, String s2) {
        int res = s1.compareToIgnoreCase(s2);
        return (res != 0)? res : s1.compareTo(s2);
      }
    }
    );
    return keys;
  }

  public boolean isModifiable(String name) {
    Style style = styles.get(name);
    if (style != null) {
      return style.isModifiable();
    }
    return true;
  }
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "jumplab" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
